@@ -11,6 +11,6 @@
 #define TAB_LOADSITE 0x30 // Only check saved names
 
 int tabPressed(char **line, size_t *line_pos, pftp_server_t ftp, 
-	       unsigned int mask);
+	       unsigned int mask, int TERM_width);
 
 #endif /* TAB_COMP_H */
