@@ -1,17 +1,26 @@
 /* -*- pftp-c -*- */
 #ifdef HAVE_CONFIG_H
-#include "config.h"
+# include "config.h"
+#endif
+#if HAVE_STDLIB_H
+# include <stdlib.h>
+#endif
+#if HAVE_MATH_H
+# include <math.h>
 #endif
-#include <stdlib.h>
-#include <math.h>
 #if HAVE_STRING_H
 # include <string.h>
 #endif
 #if HAVE_STRINGS_H
 # include <strings.h>
 #endif
+
 #include "nice_list.h"
 
+#ifdef WITH_DMALLOC
+# include <dmalloc.h>
+#endif
+
 void calc_columns(const size_t *len, size_t lens, size_t width, size_t spacing,
 		  size_t *columns, size_t **col_width, size_t *lines)
 {
