@@ -1,10 +1,16 @@
 /* -*- pftp-c -*- */
 #ifdef HAVE_CONFIG_H
-#include "config.h"
+# include "config.h"
+#endif
+#if HAVE_STDLIB_H
+# include <stdlib.h>
+#endif
+#if HAVE_STDIO_H
+# include <stdio.h>
+#endif
+#if HAVE_UNISTD_H
+# include <unistd.h>
 #endif
-#include <stdlib.h>
-#include <stdio.h>
-#include <unistd.h>
 #if HAVE_STRING_H
 # include <string.h>
 #else
@@ -12,6 +18,9 @@
 #  include <strings.h>
 # endif
 #endif
+#if HAVE_LIMITS_H
+# include <limits.h>
+#endif
 #if HAVE_DIRENT_H
 # include <dirent.h>
 #else
@@ -27,16 +36,28 @@
 #  endif
 # endif
 #endif
-#include <limits.h>
+#if HAVE_INTTYPES_H
+# include <inttypes.h>
+#endif
+#if HAVE_PWD_H
+# include <pwd.h>
+#endif
+#if HAVE_ERRNO_H
+# include <errno.h>
+#endif
+#include <time.h>
 #include <readline/readline.h>
-#include <inttypes.h>
-#include <pwd.h>
-#include <errno.h>
 
 #if !HAVE_D_TYPE
 # include <sys/stat.h>
 #endif
 
+#ifdef DEBUG
+# include <assert.h>
+#else
+# define assert(x) // x
+#endif
+
 #ifdef WITH_DMALLOC
 # include <dmalloc.h>
 #endif
@@ -50,8 +71,7 @@
 
 /* Vars from ftp_basic_cli */
 extern const command_t commands[];
-extern const int no_commands;
-extern int TERM_width;
+extern const size_t no_commands;
 extern size_t open_list_len;
 extern struct open_list *open_list;
 /* End of vars */
@@ -61,7 +81,9 @@ static void restore_localpath(char **old);
 static int fix_remotepath(const char *pos, size_t pos_len, pftp_server_t ftp, 
 			  char **old);
 static void restore_remotepath(pftp_server_t ftp, char **old);
+#ifndef WIN32
 static void add_users(tab_list_t list, const char *pos, size_t minlen);
+#endif
 static void add_commands(tab_list_t list, const char *pos, size_t minlen);
 static void add_current_sitenames(tab_list_t list, const char *pos, 
 				  size_t minlen);
@@ -73,34 +95,44 @@ static void add_remotedirs(tab_list_t list, pftp_server_t ftp, const char *pos,
 			   size_t minlen);
 static void add_remotefiles(tab_list_t list, pftp_server_t ftp, 
 			    const char *pos, size_t minlen);
-static void matches(const char *pos, size_t pos_len, tab_list_t *list);
+static void matches(const char *pos, size_t pos_len, tab_list_t *list,
+		    int *was_a_full_match);
 static void beep(void);
 static void updateline(char **line, size_t start, size_t len, 
 		       const char *match, size_t match_len);
-static void show_list(tab_list_t list);
+static void show_list(tab_list_t list, int TERM_width);
 static void get_word(const char *line, char **word, size_t line_pos, 
 		     size_t *word_start, size_t *word_end, int *quoted);
 static void str_replace(const char *needle, const char *replace, 
 			char **haystack, size_t *haystack_len);
 static void remove_dupes(tab_list_t list);
 static int alphasort2(const void *a, const void *b);
+#ifdef WIN32
+static void replace_backslash_with_frontslash(char *path);
+#endif
 
 /* tabPressed. The main function, really has very little to do with readline.
    pftp_complete calls this one. */
 int tabPressed(char **line, size_t *line_pos, pftp_server_t ftp, 
-	       unsigned int mask)
+	       unsigned int mask, int TERM_width)
 {
     char *pos = NULL;
-    int ret = 0, quoted = 0;
+    int ret = 0, quoted = 0, fullmatch = 0;
     size_t pos_len = 0, len, word_start = 0, word_end = 0, word_pos;
     tab_list_t list = NULL;
     char *word = NULL;
+    size_t add_to_pos_len_at_end = 0;
     
     /* Get the "word" to complete */
     get_word((*line), &word, (*line_pos), &word_start, &word_end, &quoted);
     word_pos = (*line_pos) - word_start;
     
     list = init_list();
+
+#ifdef WIN32
+    if (mask & TAB_LOCAL)
+	replace_backslash_with_frontslash(word);
+#endif
     
     if (mask & TAB_COMMAND) {
 	/* Get the word we want to complete */
@@ -124,12 +156,31 @@ int tabPressed(char **line, size_t *line_pos, pftp_server_t ftp,
 	char *old = NULL, *path = NULL;
 	size_t path_len = word_end - word_start;
 
-	if (!quoted)
+	if (!quoted) {
+	    size_t old;
+	    old = path_len;
 	    str_replace("\\ ", " ", &word, &path_len);    
-	
+	    add_to_pos_len_at_end = old - path_len;
+	}
+
+#ifdef WIN32
+	if (path_len >= 3 && word[1] == ':' && word[2] == '/') {
+	    path = malloc(path_len + 2);
+	    memcpy(path, word, 3);
+	    path[3] = '/';
+	    memcpy(path + 4, word + 3, path_len - 2);
+/*	    path[path_len + 1] = '\0'; */
+	} else {
+	    path = malloc(path_len + 1);
+	    memcpy(path, word, path_len + 1);
+/*	    path[path_len] = '\0'; */
+	}
+#else
 	path = malloc(path_len + 1);
 	memcpy(path, word, path_len + 1);
-	path[word_pos] = '\0';
+/*	path[path_len] = '\0'; */
+#endif
+
 	if (!(pos = strrchr(path, '/')))
 	    pos = path;
 	if (pos > path) {
@@ -147,7 +198,7 @@ int tabPressed(char **line, size_t *line_pos, pftp_server_t ftp,
 	    pos_len = old - pos;
 	else
 	    pos_len = strlen(pos);
-	
+
 	/* Read items from FS */
 	
 	if (mask & TAB_LOCAL) {
@@ -155,9 +206,13 @@ int tabPressed(char **line, size_t *line_pos, pftp_server_t ftp,
 	    path_len = strlen(path);
 
 	    if (!fix_localpath(path, path_len, &old)) {
+#ifndef WIN32
 		if (pos_len && pos[0] == '~' && path_len == 0) {
 		    add_users(list, pos, pos_len);
 		} else {
+#else
+		{
+#endif
 		    if (mask & TAB_DIRECTORY)
 			add_localdirs(list, pos, pos_len);
 		    if (mask & TAB_FILE) 
@@ -179,10 +234,13 @@ int tabPressed(char **line, size_t *line_pos, pftp_server_t ftp,
 	
 	free(path);
     }
-    
+
     /* Look for matches */
-    matches(pos, pos_len, &list);
+    matches(pos, pos_len, &list, &fullmatch);
     len = len_list(list);
+
+    /* To fix diff when removing all "\\ " */
+    pos_len += add_to_pos_len_at_end;
     
     switch (len) {
     case 0:
@@ -195,19 +253,24 @@ int tabPressed(char **line, size_t *line_pos, pftp_server_t ftp,
 	char *match = NULL;
 	size_t len = 0;
 	get_matchdata(get_first(list), &match, &len);
-	
+
 	/* Fix spaces in unquoted path's */
 	if (!quoted && (mask & (TAB_DIRECTORY | TAB_FILE))) {
-	    if (match[len-1] != ' ') {
-		str_replace(" ", "\\ ", &match, &len);
+	    if (fullmatch) {
+		if (match[len-1] != ' ') {
+		    str_replace(" ", "\\ ", &match, &len);
+		} else {
+		    match[--len] = '\0';
+		    str_replace(" ", "\\ ", &match, &len);
+		    match = realloc(match, len + 2);
+		    match[len++] = ' ';
+		    match[len] = '\0';
+		}
 	    } else {
-		match[--len] = '\0';
 		str_replace(" ", "\\ ", &match, &len);
-		match[len++] = ' ';
-		match[len] = '\0';
 	    }
 	}
-	
+
 	updateline(line, word_start + (pos - word), pos_len, match, len);
 	(*line_pos) = word_start + (pos - word) + len;
 	ret = 1;
@@ -215,7 +278,7 @@ int tabPressed(char **line, size_t *line_pos, pftp_server_t ftp,
     }; break;  
     default:
 	/* Multiple matches found, show them */
-	show_list(list);
+	show_list(list, TERM_width);
 	ret = 0;
     }
     
@@ -246,7 +309,7 @@ void get_word(const char *line, char **word, size_t line_pos,
 	if (!pos)
 	    break;
 	
-	if (pos - line > line_pos)
+	if ((pos - line) > line_pos)
 	    break;
 	
 	open = 1;
@@ -274,10 +337,15 @@ void get_word(const char *line, char **word, size_t line_pos,
 	}
 	
 	(*word_start) = c;
-	if ((pos = strchr(line + (*word_start), ' ')))
-	    (*word_end) = pos - line;
-	else
-	    (*word_end) = (*word_start) + strlen(line + (*word_start));    
+	if ((pos = strchr(line + (*word_start), ' ')) 
+	    && ((pos - line) < line_pos)) {
+	    (*word_end) = (*word_start) + strlen(line + (*word_start));
+	} else {
+	    if ((pos = strchr(line + (*word_start), ' ')))
+		(*word_end) = pos - line;
+	    else
+		(*word_end) = (*word_start) + strlen(line + (*word_start));    
+	}
     }
     
     (*word) = malloc(((*word_end) - (*word_start)) + 1);
@@ -285,6 +353,7 @@ void get_word(const char *line, char **word, size_t line_pos,
     (*word)[(*word_end) - (*word_start)] = '\0';
 }
 
+#ifndef WIN32
 /* adds all usernames longer or the same as minlen to the list.
    Only called if pos[0] == '~' */
 void add_users(tab_list_t list, const char *pos, size_t minlen)
@@ -318,6 +387,7 @@ void add_users(tab_list_t list, const char *pos, size_t minlen)
 
     if (tmp) free(tmp);
 }
+#endif
 
 /* adds all commands longer or the same as minlen to the list */
 void add_commands(tab_list_t list, const char *pos, size_t minlen)
@@ -536,7 +606,8 @@ void add_remote(tab_list_t list, pftp_server_t ftp, size_t minlen,
     if (append) {
 	for (f = 0; f < names; f++) {
 	    len = strlen(name[f]);
-	    if (len >= minlen && strcmp(name[f], ".")) {	
+	    if (len >= minlen 
+		&& strcmp(name[f], ".") && strcmp(name[f], "..")) {	
 		name[f][len] = append;
 		add_item(list, name[f], len + 1);
 	    }
@@ -545,7 +616,8 @@ void add_remote(tab_list_t list, pftp_server_t ftp, size_t minlen,
     } else {
 	for (f = 0; f < names; f++) {
 	    len = strlen(name[f]);
-	    if (len >= minlen && strcmp(name[f], ".")) {	
+	    if (len >= minlen 
+		&& strcmp(name[f], ".") && strcmp(name[f], "..")) {	
 		add_item(list, name[f], len);
 	    }
 	    free(name[f]);
@@ -577,12 +649,15 @@ void add_remotefiles(tab_list_t list, pftp_server_t ftp,
     add_remote(list, ftp, minlen, sel_rfile, pos[minlen] != ' ' ? ' ' : 0);
 }
 
-void int_matches(const char *pos, size_t pos_len, 
-		 tab_list_t list, tab_list_t bak, 
-		 int (*cmpnstrfunc)(const char *, const char *, size_t len),
-		 int (*cmpcharfunc)(char, char), int casesensitive)
+static void int_matches(const char *pos, size_t pos_len, 
+			tab_list_t list, tab_list_t bak, 
+			int (*cmpnstrfunc)(const char *, const char *, 
+					   size_t len),
+			int (*cmpcharfunc)(char, char), int casesensitive,
+			int *was_a_fullmatch)
 {
     tab_item_t item = get_first(list);
+    *was_a_fullmatch = 1;
     
     while (item) {
 	if (cmpnstrfunc(pos, get_match(item), pos_len)) {
@@ -600,6 +675,8 @@ void int_matches(const char *pos, size_t pos_len,
 	int same_case = 1, stop;
 	tab_item_t item2;
 
+	*was_a_fullmatch = 0;
+
 	memset(build_new, 0, new_buf);
 	
 	if (casesensitive) {
@@ -687,11 +764,13 @@ int charcasecmp(char c1, char c2)
 }
 
 /* Look's for matching completions in list */
-void matches(const char *pos, size_t pos_len, tab_list_t *list)
+void matches(const char *pos, size_t pos_len, tab_list_t *list, 
+	     int *was_a_fullmatch)
 {
     tab_list_t bak = init_list();
     
-    int_matches(pos, pos_len, (*list), bak, strncmp, charcmp, 0); 
+    int_matches(pos, pos_len, (*list), bak, strncmp, charcmp, 0, 
+		was_a_fullmatch); 
     
     if (len_list((*list)) > 0) {
 	/* Found any with same case? Go for it. */
@@ -704,7 +783,8 @@ void matches(const char *pos, size_t pos_len, tab_list_t *list)
     move_list(list, bak);
     bak = init_list();
 
-    int_matches(pos, pos_len, (*list), bak, strncasecmp, charcasecmp, 1); 
+    int_matches(pos, pos_len, (*list), bak, strncasecmp, charcasecmp, 1,
+		was_a_fullmatch); 
     
     free_list(bak);
 }
@@ -720,18 +800,19 @@ void beep(void)
 void updateline(char **word, size_t start, size_t len, const char *match,
 	    size_t match_len)
 {
+    size_t l;
+
     /* match_len will never be less than len (or it deserves to crash) */
-    if (match_len > len) {
-	size_t l = (strlen((*word)) + 1) - len;
-	(*word) = realloc((*word), l + match_len);
-	memmove((*word) + start + match_len, (*word) + start + len, l - start);
-    }
-    
+    assert(match_len >= len);
+
+    l = (strlen((*word)) + 1) - len;
+    (*word) = realloc((*word), l + match_len);
+    memmove((*word) + start + match_len, (*word) + start + len, l - start);
     memcpy((*word) + start, match, match_len);
 }
 
 /* Show list of possible completions */
-void show_list(tab_list_t list)
+void show_list(tab_list_t list, int TERM_width)
 {
     size_t count = len_list(list);
     size_t l = 0, *len = malloc(count * sizeof(size_t));
@@ -757,7 +838,8 @@ void show_list(tab_list_t list)
 	    if ((l = y+(x*lines)) >= count)
 		break;
 	    fwrite(match[l], 1, len[l], stdout);
-	    fwrite(space, 1, col_width[x] - len[l], stdout);
+	    if (col_width[x] > len[l])
+		fwrite(space, 1, col_width[x] - len[l], stdout);
 	    free(match[l]);
 	}
 	puts("");
@@ -871,13 +953,13 @@ void str_replace(const char *needle, const char *replace, char **haystack,
 	while ((pos = strstr(start, needle))) {
 	    memcpy(pos, replace, r_l);
 	    memmove(pos + r_l, pos + n_l, 
-		    (*haystack_len)-((pos+n_l)-(*haystack)));
+		    (*haystack_len) - ((pos + n_l) - (*haystack)));
 	    (*haystack_len) -= (n_l - r_l);
 	    (*haystack)[(*haystack_len)] = '\0';
 	    start = pos + r_l;
 	}
     } else if (n_l < r_l) {
-	/* Reallocation and moving needed */
+	/* Reallocation and moving needed */	
 	size_t p;
 	while ((pos = strstr(start, needle))) {
 	    p = pos - (*haystack);
@@ -919,3 +1001,16 @@ void remove_dupes(tab_list_t list)
 	item1 = get_nextitem(item1);
     }
 }
+
+#ifdef WIN32
+void replace_backslash_with_frontslash(char *path)
+{
+    char *p;
+    p = path;
+    while (*p) {
+	if (*p == '\\')
+	    *p = '/';
+	p++;
+    }
+}
+#endif
